REBOL [
	Title: "RSP handler"
	Author: "SOFTINNOV / Nenad Rakocevic"
	Version: 1.0.0
	Date: 27/05/2007
]

do-cache uniserve-path/libs/html.r
do-cache uniserve-path/libs/headers.r
do-cache uniserve-path/libs/decode-cgi.r
do-cache uniserve-path/libs/url.r

if exists?-cache file: %misc/rsp-init.r [
	do-cache file
	if value? 'on-rsp-start [attempt [on-rsp-start]]	;TODO: expose errors !!
]

databases: none

file: %httpd.cfg
all [
	any [
		all [
			exists? file
			block? conf: load/all file
		]
		all [
			exists?-cache file
			block? conf: load-cache file
		]
	]
	parse conf [thru 'globals into [thru 'databases set databases block!]]
]

install-module [
	name: 'RSP

	verbose: 0
	
	apps: make block! 1 ; [app block! ...]
	jobs: make block! 1
	start-flag: close-flag: no
	splitted: none
	
	evt-class: context [
		on-application-start: 
		on-application-end: 
		on-session-start: 
		on-session-end: 
		on-page-start:
		on-page-end: 
		none
	]

	save-path: copy system/options/path
	set 'halt set 'q set 'quit does [throw 'force-quit]

	set 'rsp-log func [v] compose [(:print) mold v v]
	rsp-prin: func [data][insert tail response/buffer reform data data]
	rsp-print: func [data][
		insert tail response/buffer reform data
		insert tail response/buffer newline
	]
	rsp-probe: func [data][insert tail response/buffer mold data data]
	rsp-print-funcs: reduce [:rsp-prin :rsp-print :rsp-probe]
	saved-print-funcs: reduce [:prin :print :probe]
	print-funcs: [prin print probe]

	form-error: func [err [object!] /local type id desc][
		type: err/type
		id: err/id
		arg1: err/arg1
		arg2: err/arg2
		arg3: err/arg3
		desc: reduce system/error/:type/:id
		print {
<HTML>
	<HEAD>
		<TITLE>
			RSP Error
		</TITLE>
	</HEAD>
	<BODY><FONT face="Arial">
		<BR>
		<CENTER>
		<H2>&gt; RSP Error Trapped&lt;</H2><BR>
		<TABLE border="1" cellspacing="0" cellpadding="0">
		<TR><TD align="right"><FONT face="Arial"><B>Error Code :</B></FONT></TD>
		<TD align="left"><FONT face="Arial">}print mold err/code print{</FONT></TD></TR>
		<TR><TD align="right"><FONT face="Arial"><B>Description :</B></FONT></TD>
		<TD align="left"><FONT face="Arial">}print ["<I>"type " error ! </I><BR>"desc] print{</FONT></TD></TR>
		<TR><TD align="right"><FONT face="Arial"><B>Near :</B></FONT></TD>
		<TD align="left"><FONT face="Arial">}print html-encode mold/flat err/near print{</FONT></TD></TR>
		<TR><TD align="right"><FONT face="Arial"><B>Where :</B></FONT></TD>
		<TD align="left"><FONT face="Arial">}print html-encode mold/flat err/where print{</FONT></TD></TR>
		</TABLE>
		</CENTER>
	</BODY>
</HTML>
			}	
	]
	
	protected-exec: func [code [block! function!] /local res thru?][
		if not thru?: :print = :rsp-print [
			set print-funcs rsp-print-funcs	
		]
		if error? set/any 'res try [
			catch any [all [block? :code code][code]]
		][	
			form-error disarm :res
			set print-funcs saved-print-funcs
			return false
		]
		if not thru? [set print-funcs saved-print-funcs]
		true
	]
	
	engine: context [
		list: make block! 16
		current: none
		verbose: 0
		
		sandbox: context [
			txt: func [s o][
				insert/part tail response/buffer at current s o
			]
			cat: []
		]
		
		compile: func [entry /no-lang /local out value s e word id][
			if not no-lang [
				id: locale/lang
				locale/set-default-lang
			]
			out: make string! 1024			
			parse/all current: fourth entry [
				any [
					end break
					| "#[" copy value to #"]" skip (					
						append out reform [
							" prin any [pick cat"
							locale/id? value
							mold value #"]"
						]
					)
					| "<%" [#"=" (append out " prin ") | none]
						copy value [to "%>" | none] 2 skip (
							if value [repend out [value #" "]]
						)
					| s: copy value [any [e: "<%" :e break | e: "#[" :e break | skip]] e: (
						append out reform [" txt" index? s offset? s e #" "]
					)
				]
			]
			if not no-lang [locale/set-lang id]
			if error? try [out: load out][			
				out: reduce ['load out]
			]		
			poke entry 3 out: bind out in sandbox 'txt		
			out
		]

		refresh: func [entry file][
			poke entry 4 read file
			compile entry
		]

		add-file: func [path file /local pos][
			repend list [path modified? file none read file]
			skip tail list -4
		]
		
		dump: has [cnt][
			cnt: 1
			foreach [a b c d] list [
				print [a b]
				print ["code:" mold c]
				cnt: cnt + 1
			]
		]

		exec: func [path file /local code pos][
			if verbose > 0 [
				rsp-log reform ["executing file: " mold file "path: " mold path]
			]				
			repend jobs [path current]
			code: either pos: find list path [
				if pos/2 < modified? file [refresh pos file]
				current: fourth pos				
				third pos
			][
				compile add-file path file
			]		
			if verbose > 1 [rsp-log reform ["code :" mold code]]
			protected-exec code
			current: last jobs
			clear skip tail jobs -2
			if verbose > 2 [dump]
		]
	]
		
	debug: context [
		active?: none
		menu-head: read-cache %misc/debug-head.html
		menu-code: read-cache %misc/debug-menu.rsp
		menu: reduce [none 01/01/3000 none menu-code]
		engine/compile/no-lang menu
		
		insert-menu: has [buf pos body][
			if not active? [exit]
			buf: copy response/buffer
			clear response/buffer
			protected-exec menu/3
			either pos: find buf "</head>" [
				insert pos menu-head
			][
				if body: find buf "<body" [
					insert body "^/<head>^/"
					insert body menu-head
					insert body "^/</head>^/"
				]
			]
			if any [body body: find buf "<body"] [
				insert find/tail body ">" response/buffer
			]
			clear response/buffer
			insert response/buffer buf
		]
		
		make-redirect-page: func [url][
			clear response/buffer
			append response/buffer {
				<html>
					<head></head>
					<body><font face="Arial"><center>
						<h2>Redirection Catched</h2>
						<br><br>Destination URL: }
			append response/buffer rejoin [
				{<a href="} url {">} url
				"</a></center></font></body></html>"
			]
		]
	]
	
	;--- public API ---
	
	set-protected: func [w [word!] value][
		unprotect :w
		set :w :value
		protect :w
	]
	
	set 'locale context [
		path: lang:	current: default: none
		cache: make block! 1
		alpha: charset [#"A" - #"Z" #"a" - #"z"]

		set-lang: func [id [word! none!] /local cats file][
			if not all [path id][return true]
			if not cats: select cache path [
				repend cache [path cats: make block! 1]
			]
			if not current: select cats id  [
				file: join request/config/root-dir [
					slash path mold id slash mold id ".cat"
				]
				if not exists? file [return none]
				repend cats [id current: make hash! load file]
			]
			lang: id
			engine/sandbox/cat: current
			true
		]
		get-path: does [
			join request/config/root-dir [slash path mold lang slash]
		]
		set 'say func [[catch] data [string! none!] /local cat][	
			if not all [current default data][return data]				
			any [
				all [
					cat: find/case default data
					cat: pick current index? cat
					cat
				]
				data
			]
		]
		id?: func [txt [string!] /local pos][		
			any [
				all [
					default
					pos: find/case default txt
					index? pos
				]
				0
			]
		]
		set-default-lang: does [
			set-lang any [select request/config 'default-lang 'en]
			default: current
		]
		decode-lang: has [id list v v2][
			if not path: select request/config 'locales-dir [exit]
			set-default-lang
			current: none
			if all [
				session/active?
				word? id: select session/content 'lang
			][
				if set-lang id [exit]
			]
			if id: select request/headers 'Accept-language [
				list: make block! 1
				parse id [
					some [
						copy v 1 8 alpha (append list v)
						opt [
							"-" copy v2 1 8 alpha
							(insert back tail list rejoin [v #"-" v2])
						]
						opt ","
					]
					";" thru end
				]
				foreach id list [
					if set-lang to word! id [exit]
				]
			]
		]
	]

	set 'do-sql func [
		[catch] db [word!] data [string! block! word!]
		/flat /local port out res
	][
		if not block? databases [
			throw make error! "no database defined!"
		]
		if not pos: find databases :db [
			throw make error! rejoin ["database" :db "not defined!"]
		]
		port: pos/2
		either error? set/any 'res try [
			if any [block? port url? port][				
				poke pos 2 port: open port				
				if not port/handler [	;-- RT's drivers
					poke pos 2 port: first port
				]
			]
		
			either word? data [
				db-cache/query db data flat
			][
				poke response/stats 1 response/stats/1 + 1
				either value? 'send-sql [
					either flat [
						send-sql/flat port data
					][
						send-sql port data
					]
				][
					either port/handler [
						any [
							insert port data
							copy port
						]
					][							;-- RT's drivers
						insert port data
						either flat [
							out: make block! 8
							until [if data: pick port 1 [append out data] data]
							out
						][
							attempt [copy port]
						]
					]
				]
			]
		][
			throw res
		][
			either value? 'res [res][none]
		]
	]
	
	set 'db-cache context [
		data: make block! 1  ; ['db ['class [...] ...] ...]
		sync-list: none

		sync: func [list [block! none!] /local times][	
			if not all [none? list sync-list][
				sync-list: list
			]
		]
		
		define: func [db [word!] spec [block!] /local pos times][
			parse spec [
				some [
					pos: string! (
						change/only pos reduce [01/01/0001 pos/1 make block! 16]
					)
					| skip
				]
			]
			repend data [db spec]
			if none? sync-list [sync-list: make block! length? spec]
			times: make block! 1
			loop (length? spec) / 2 [append times now]
			repend sync-list [db times]
		]
		
		query: func [
			db [word!] class [word!] flat [logic! none!]
			/local cache pos modified
		][
			cache: second pos: find data/:db class			
			modified: pick sync-list/:db divide 1 + index? pos 2
			if cache/1 <> modified [
				cache/1: modified
				clear cache/3
				append cache/3 either flat [
					do-sql/flat db cache/2
				][
					do-sql db cache/2
				]
			]
			copy cache/3
		]
		
		invalid: func [db [word!] class [word! block!] /local cache list][
			if word? class [class: reduce [class]]
			cache: data/:db
			list: sync-list/:db
			foreach w class [
				poke list (1 + index? find cache w) / 2 now
			]
		]
	]
	
	set 'request context [
		content: headers: method: posted: client-ip: server-port: 
		translated: parsed: config: web-app: none
		
		query-string: has [out][
			out: make string! 32
			foreach-nv-pair params [
				insert tail out name
				insert tail out #"="
				insert tail out url-encode value
				insert tail out #"&"
			]
			remove back tail out			
			out
		]
	]
	
	set 'response context [
		buffer: make string! 1024 * 64
		headers: status: forward-url: none
		buffered?: yes
		cache-list: make block! 1
		stats: [0]	; [sql-requests]
		
		set-status: func [code [integer!] /msg str [string!]][
			status: any [all [msg reform [code str]] code]
		]
		set-cookie: func [][]
		set-header: func [word [word!] value [string! none!] /add /local pos][
			if not headers [headers: make block! 1]
			either all [not add pos: find headers word][
				change next pos value
			][
				insert tail headers word
				insert tail headers value
			]
		]
		end: does [throw 'force-quit]
		reset: does [clear buffer]
		redirect: func [url [string! url!] /temp][
			either debug/active? [
				debug/make-redirect-page url
			][
				set-status any [all [temp 302] 301]
				set-header 'Location form url
			]
			end
		]
		auto-flush: func [mode [logic!]][
			;--TDB
		]
		flush: func [/end][
			if buffered? [buffered?: no]
			if all [not end empty? buffer][exit]		; don't send a false 0-length body			
			server-send-data mold/all compose/only/deep [
				part [content (buffer) headers (headers)]
			]
			reset
		]
		cache: func [time [time!]][insert tail cache-list time]
		cache-delete: func [url [url!]][insert tail cache-list time]
		forward: func [[catch] url [string! url!]][
			if any [
				all [string? url slash <> first url]
				all [url? url not get in parse-url url 'target]
			][			
				throw make error! "invalid URL argument"
			]
			if not buffered? [
				throw make error! "forward cannot be used after flush"
			]
			if not empty? buffer [clear buffer]
			forward-url: url
			end
		]
	]
	
	set 'session context [
		content: timeout: events: id: none
		active?: init?: no
		
		add: func [name [word!] value /local pos][
			either pos: find content name [
				change next pos value
			][
				repend content [name value]
			]
		]
		
		remove: func [name [word!]][
			system/words/remove/part find content name 2
		]
		
		exists?: func [name [word!]][
			to logic! find content name
		]
		
		start: does [
			if not active? [
				start-flag: yes
				close-flag: no
				content: make block! 1
				id: none
			]
			true
		]
		
		reset: does [
			if active? [
				;change-dir request/config/root-dir
				if exists? 'login? [content/login?: no]
				fire-event 'on-session-start
				;change-dir save-path
			]
		]
		
		end: does [
			start-flag: no
			close-flag: yes	
		]
	]
	
	set 'include func [file [file!] /local blk path count][
		count: [0]
		either count/1 > 5 [
			print "### Include RSP failed: max inclusion level reached! ###"
		][
			count/1: count/1 + 1
			path: first splitted
			either slash = first file [
				blk: split-path file 
				engine/exec join path first blk second blk
			][
				engine/exec join path file file
			]
			count/1: count/1 - 1
		]		
		if all [
			integer? response/status
			response/status >= 300
		][
			response/end
		]
	]
	set 'include-file func [file [file!]][
		insert tail response/buffer read file
	]
	
	set 'validate func [[catch] spec [block!] /full /local vars value invalid pos m?][
		full: to logic! full
		if error? try [
			vars: request/content
			value: pick [[name type m?][name type]] full
			foreach :value spec [
				either pos: find vars name [
					either empty? value: pick pos 2 [
						poke pos 2 value: none
					][
						if all [
							type <> '-
							not attempt [poke pos 2 to get type value]
						][
							if none? invalid [invalid: make block! 1]
							insert tail invalid name
						]
					]
				][
					insert tail vars name
					insert tail vars value: none
				]
				if all [none? :value m? m? = '*][
					if none? invalid [invalid: make block! 1]
					insert tail invalid name
				]
			]
		][throw make error! "invalid spec block!"]
		invalid
	]
	
	protect [
		do-sql db-cache request response session include
		include-file validate locale say
	]
	
	;--- end of public API ---
	
	set 'lf #"^/"
	nl: [crlf | cr | lf]
	crlfx2: join crlf crlf
	dquote: #"^""
	
	reset-response-object: does [
		clear response/cache-list
		response/status: response/forward-url: none
		response/buffered?: yes
		if response/headers [clear response/headers]
		if empty? jobs [response/reset]
	]
	
	;-- quick implementation of multipart decoding :
	;	- doesn't support multipart/mixed encoding yet
	;	- doesn't parse all optional headers
	
	decode-multipart: func [data /local bound list name filename value pos][
		parse/all request/headers/Content-type [
			thru "boundary=" opt dquote copy bound [to dquote | to end]
		]
		if not bound [return ""]	 ;-- add proper error handler
		insert bound "--"	
		list: make block! 2
		parse/all data [
			some [
				bound nl some [
					thru {name="} copy name to dquote skip
					[#";" thru {="} copy filename to dquote | none]
					thru crlfx2 copy value to bound (
						clear back back tail value ; -- delete ending crlf (watch out for cr or lf only!!)
						if all [
							#"%" = first value
							".tmp" = skip tail value -4
						][
							value: load value
						]
						insert tail list name
						either filename [						
							insert/only tail list reduce [filename value]
						][
							insert tail list value
						]
						filename: none
					) | "--"
				]
			]
		]		
		list
	]
	
	decode-params: has [list value pos][
		list: make block! 1
		any [
			all [
				value: select request/headers 'Content-type
				find/part value "multipart/form-data" 19
				list: decode-multipart request/posted
			]
			all [
				request/posted
				list: decode-cgi/raw/with request/posted list
			]
		]	
		if value: request/parsed/arg [
			if pos: find/last value #"#" [
				clear at value: copy value index? pos
			]						
			list: decode-cgi/raw/with value list
		]		
		request/content: list
	]
	
	decode-msg: func [data /local value list init?][
		parse load/all data [
			'cfg  set value block! 	 (request/config: value)
			'in	  set value object!	 (request/parsed: value)
			'ip   set value tuple!	 (request/client-ip: value)
			'port set value integer! (request/server-port: value)
			'session [
				into [
					(session/active?: yes)
					'ID set value string! (session/id: value)
					'vars set list block! (session/content: list)
					'queries set list [block! | none!] (db-cache/sync list)
					'app  set value [string! | none!]  (request/web-app: value)
					'timeout set value time! (session/timeout: value)
					'init set value logic! (session/init?: value)
				]
				| none! (
					session/active?: session/init?: no
					session/content: session/id: none
					request/web-app: none
				)
			]
		]
		request/method: request/parsed/method
		request/posted: request/parsed/content
		request/headers: request/parsed/headers
		request/translated: join request/config/root-dir [
			request/parsed/path
			request/parsed/target
		]				
		if find request/config 'debug [debug/active?: yes]
	]
	
	fire-event: func [event [word!]][
		protected-exec get in session/events :event
	]
	
	process-events: has [pos evt-data events][	
		either request/web-app [
			either events: select apps request/web-app [
				session/events: events
			][
				evt-data: any [
					attempt [load join request/config/root-dir %/app-init.r] ; TBD: report errors !!!
					[]
				]				
				repend apps [
					request/web-app
					events: make evt-class evt-data
				]
				pos: skip tail apps -3
				session/events: events
				change-dir request/config/root-dir
				fire-event 'on-application-start		 ; TBD: report errors on all events
				system/script/path: dirize save-path
			]			
			if session/init? [
				change-dir request/config/root-dir				
				fire-event 'on-session-start			; TBD: report errors on all events
				change-dir save-path
			]
		][
			session/events: none
		]
	]
	
	build-msg: has [res sess][	
		res: compose/only [
			status	(response/status)
			headers	(response/headers)
			content (response/buffer)
			forward (response/forward-url)
			session
		]
		either any [session/active? start-flag][
			sess: compose/only [
				id	 (session/id)
				vars (session/content)
				queries (db-cache/sync-list)
				timeout (session/timeout)
			]
			if start-flag [insert sess 'init]
			if close-flag [append sess 'close]
			append/only res sess
		][
			append res none
		]
		mold/all res
	]

	on-task-received: func [data /local file events][
		if verbose > 0 [
			print "New job received" 
			print mold data
		]
		session/events: none
		close-flag: start-flag: no
		debug/active?: no

		decode-msg data		
		process-events
		decode-params
		locale/decode-lang
		reset-response-object

		file: request/translated
		if verbose > 0 [print rejoin ["file: " mold file]]
		change-dir first splitted: split-path file
				
		if session/events [fire-event 'on-page-start]
		engine/exec file last splitted	
		if session/events [fire-event 'on-page-end]
		
		if verbose > 2 [print mold response/buffer]
		if not empty? response/buffer [debug/insert-menu]
		change-dir save-path
		
		all [
			not response/buffered?
			not empty? response/buffer
			response/flush
			response/flush/end
		]		
		result: build-msg 
	]
	
	on-quit: does [
		foreach [app events] apps [attempt [events/on-application-end]]
		if value? 'on-rsp-end [attempt [on-rsp-end]]
		if block? databases [
			parse databases [any [set p port! (attempt [close p]) | skip]]
		]
	]
]
