REBOL []

install-HTTPd-extension [
	name: 'mod-extapp
	verbose: 0
	
	extapp!: context [name: childs: jobs: balancing: specs: none]
	app!:    context [pid: ports: extapp: none]

	templates: make block! 1 ; [[spec] ...]
	actives: make block! 1 ; [name extapp! ...]
	;-- extapp/jobs specs: [id client req ...]

	launch-servers: has [cmd ret new n v][
		foreach spec templates [
			parse spec [
				some [
					'environment into [
						any [set n skip set v skip (set-env form n form v)]
					]
					| 'command set cmd string!
					| skip
				]
			]
			new: make extapp! [
				name: spec/name
				jobs: make block! 8
				childs: make block! 1
				balancing: 'round-robin
				specs: spec
			]
			if cmd [
				ret: launch-app cmd
				either ret/1 = 'ok [
					new/childs: all [
						cmd ret/1 = 'ok reduce [
							make app! [
								pid: ret/2
								specs: spec
								server: new
							]
						]
					]
				][
					log/error reform [
						"cannot launch :" cmd newline 
						"OS message:" trim form ret/2
					]
				]
			]
			repend actives [spec/name new]
		]
	]
	
	kill-servers: does [
		foreach [name extapp] actives [
			foreach app extapp/childs [
				if app/pid [
					kill-app app/pid
					if verbose > 0 [log/info reform [name "killed"]]
				]
			]
		]
	]
	
	set 'extapp-register func [
		scheme [word!]
		evt [block!]
		/with defs [block!]
		/local out url ctx port delay
	][
		out: make block! 1
		foreach [name extapp] actives [
			if all [
				url: select extapp/specs 'url
				scheme = get in parse-url url 'scheme
			][
				if number? delay: select extapp/specs 'delay [
					wait delay					
				]
				foreach app extapp/childs [
					ctx: service/client				
					defs: repend any [defs make block! 2][
						to-set-word 'jobs extapp/jobs
					]
					port: open-port/with url evt defs
					service/set-peer ctx
					either app/ports [
						append app/ports port
					][
						app/ports: reduce [port]
					]
				]
				append out name
			]
		]
		out
	]

	set 'extapp-get-info func [
		id [word!]
		/local extapp app jobs port
	][
		all [
			any [
				extapp: select actives id
				all [print "error, extapp not found" false]
			]
			all [
				jobs: extapp/jobs
				extapp/childs
				not empty? extapp/childs
				app: first extapp/childs
			]
			all [
				app/ports
				not empty? app/ports
				port: first app/ports
			]
		]
		reduce [port jobs]
	]

	on-started: func [svc][
		launch-servers
	]
	
	on-reload: func [svc][
		kill-servers
	]
	
	on-quit: func [svc][
		kill-servers
	]
	
	words: [
		extern-app: [block!] in globals do [
			append/only templates args/1
		]
	]
]
	