REBOL [
	Title: "Uniserve: task-master service"
	Author: "SOFTINNOV / Nenad Rakocevic"
	Version: 1.3.0
	Date: 06/06/2007
	History: [
		1.3.0 - 06/06/2006 {
			o Added 'locals argument to all 'on-task-* callbacks
		}
		1.2.0 - 17/01/2006 {
			o Added 'part message support triggering the new 'on-task-part callback.
		}
		1.1.1 - 23/03/2005 {
			o Cleaner handling of errors for 'on-task-failed invocation.
		}
		1.1.0 - 01/02/2005 {
			o Client/user-data added to the saved context for each task.
		}
	]
]

install-service [
	name: 'task-master
	port-id: 9799
	;hidden: yes
	verbose: 0
	
	bg-process: %services/task-master/task-handler.r
	pool-list: make block! 50
	obj: len: none
	queue: copy []
	additional-args: ""

	share [
		pool-start: 2	; number of helper process when Uniserve starts
		pool-max: 	4	; maximum number of helper process
		pool-count: 0	; read-only - do not change it !!
		job-max: 	100	; max size of the waiting jobs queue
		do-task: func [data service /save locals][
			process-task data service service/client any [locals none]		
		]
	]

	task: context [
		busy: no
		ctx: none
	]

	fork: does [
		if verbose > 0 [log/info "launching new slave"]
		launch* join mold either value? 'uniserve-path [
			rejoin [uniserve-path slash bg-process]
		][
			bg-process
		] additional-args
		shared/pool-count: shared/pool-count + 1
	]	
	
	send-data: func [port data][
		len: debase/base to-hex length? data: mold data 16
		insert data len
		write-client/with data port
	]
	
	process-task: func [data server port locals /local list obj][
		list: :pool-list
		while [all [not tail? list list/1/user-data/busy]][list: next list]
		either tail? list [
			if any [
				zero? shared/pool-max
				shared/pool-max > shared/pool-count 
			][
				fork
			]
			either shared/job-max > length? queue  [
				if verbose > 0 [log/info "queuing job"]
				repend/only queue [copy/deep data server port locals]
			][
				if in obj: client/user-data/ctx/1 'on-task-failed [
					obj/on-task-failed 'overload
				]
			]
		][
			if verbose > 0 [
				log/info join "new task affected using module: " server/module
			]
			send-data list/1 reduce [server/name server/module data]
			list/1/user-data/busy: yes
			list/1/user-data/ctx: reduce [server port locals]
		]
	]
	
	on-started: does [
		additional-args: rejoin [
			" uniserve-path " mold uniserve-path
			" server-port "	 mold port-id
			either value? 'modules-path [
				rejoin [" modules-path " mold modules-path]
			][""]
		]
		if integer? shared/pool-start [loop shared/pool-start [fork]]
	]

	on-new-client: has [tmp][
		if client/remote-ip <> 127.0.0.1 [close-client exit]
		set-modes client [no-delay: on]
		client/timeout: 15
		client/user-data: make task []
		append pool-list :client
		stop-at: 4
		if verbose > 0 [log/info "new slave process connected"]
		if not empty? queue [
			tmp: queue/1
			send-data client reduce [tmp/2/name tmp/2/module tmp/1]
			client/user-data/busy: yes
			client/user-data/ctx: reduce [tmp/2 tmp/3 tmp/4]
			remove queue
			if verbose > 0 [log/info "new slave got job"]
		]
	]
	
	on-close-client: does [
		remove find pool-list :client
		if client/user-data/busy [
			if in obj: client/user-data/ctx/1 'on-task-failed [
				obj/on-task-failed 'error client/user-data/ctx/3
			]
		]
		if verbose > 0 [log/info "slave process closed"]
		shared/pool-count: shared/pool-count - 1
	]
	
	
	on-received: func [data /local tmp svc sav-client locals][
		either stop-at = 4 [
			if verbose > 1 [log/info "header received"]
			stop-at: to integer! copy data
		][
			if verbose > 1 [log/info "body received"]
			ctx: client/user-data/ctx
			svc: ctx/1
			sav-client: svc/client
			;sav-data: svc/client/user-data
			svc/client: svc/peer: ctx/2				; restore client port
			;if ctx/3 [svc/client/user-data: ctx/3]	; restore client local data			
			locals: ctx/3
			data: first load/all to string! data
			switch first data [
				ok		[svc/on-task-done data/2 locals]
				part	[
					if in svc 'on-task-part [svc/on-task-part data/2 locals]
					svc/client: svc/peer: sav-client
					;svc/client/user-data: sav-data
					stop-at: 4
					exit
				]
				error	[			
					if in svc 'on-task-failed [
						svc/on-task-failed first reduce next data locals
					]
				]
			]
			svc/client: svc/peer: sav-client
			;svc/client/user-data: sav-data
			either empty? queue [
				client/user-data/busy: no
				client/user-data/ctx: none
				if zero? shared/pool-max [close-client]
			][
				tmp: queue/1
				send-data client reduce [tmp/2/name tmp/2/module tmp/1]
				client/user-data/ctx: reduce [tmp/2 tmp/3 tmp/4]
				remove queue
				if verbose > 0 [log/info "no wait-state slave reuse"]
			]
			stop-at: 4
		]
	]
]