REBOL [
	Title: "Uniserve: task-master client"
	Author: "SOFTINNOV / Nenad Rakocevic"
	Version: 1.1.0
	Date: 14/07/2006
]

if ssa: system/script/args [
	ssa: parse ssa none
	uniserve-path: all [v: select ssa "uniserve-path" load v]
	modules-path:  all [v: select ssa "modules-path" load v]
	uniserve-port: all [v: select ssa "server-port" load v]
]
if not value? 'uniserve-path [uniserve-path: what-dir]
if not value? 'modules-path  [modules-path: dirize uniserve-path/modules]
if not value? 'uniserve-port [uniserve-port: 9799]

change-dir system/options/path

either value? 'encap-fs [
	do-cache uniserve-path/libs/log.r
][
	do uniserve-path/libs/encap-fs.r
	change-dir system/options/path
	do-cache uniserve-path/libs/log.r
]

s-read-io: get in system/words 'read-io
s-quit: get in system/words 'quit
s-halt: get in system/words 'halt

ctx-task-class: make log-class [
	name: 'task-handler
	verbose: 0
	
	module: server: scheme: state: err: req: t0: len: remains: none
	packet: make binary! buf-size: 1024 * 16
	request: make binary! buf-size
	
	server-address: 127.0.0.1
	server-port: uniserve-port
	
	modules: make block! 5
	
	set 'install-module func [body [block!]][
		module: make make log-class [on-task-received: result: none] body
		if verbose > 0 [log/info ["installing module : " mold module/name]]
		repend modules [module/name module]
		recycle
	]

	set 'server-send-data func [data][
		insert data debase/base to-hex length? data 16
		insert server data
	]

	connect: does [
		if error? try [
			server: open/binary/direct/no-wait rejoin [tcp:// server-address ":" server-port]
			set-modes server [no-delay: on]
		][quit]
		if all [value? 'no-screen do no-screen][
			logger/level: 'file
			logger/file.log: join %bg-process- [form server/local-port %.log]
		]
		if verbose > 0 [log/info "connected to Uniserve"]
		forever [
			clear request
			clear packet
			if verbose > 0 [log/info "waiting for task..."]
			wait server
			state: s-read-io server packet 4			
			any [
				all [
					state <= 0
					close server
					s-quit
				]
				all [
					state = 4
					remains: len: to integer! packet
				]
				all [
					state < 4
					len: to integer! copy/part packet 4
					remains: len - length? packet: at packet 5
					insert tail request packet
				]
			]
			if verbose > 0 [log/info ["head received, body:" len]]
			until [
				clear packet
				wait server
				state: s-read-io server packet min remains buf-size			
				if verbose > 1 [log/info ["state:" state]]
				if positive? state [
					remains: remains - length? packet
					insert tail request packet
				]
				any [zero? remains state <= 0]
			]
			;if negative? state [close server quit]
			if verbose > 0 [t0: now/time/precise]
			req: load to string! request
			either error? set/any 'err try [
				if not find modules req/2 [
					do-cache join modules-path [req/2 %.r]
					change-dir system/options/path
				]
				module: select modules req/2
				module/result: none			
				module/on-task-received load to string! req/3
				false
			][		
				err: disarm err				
				log/error mold err
				server-send-data remold ['error mold/all err]
				change-dir system/options/home
				write/append to-file join any [req/2 "noname"] ".log" reform [
					newline now ":" mold :err
				]
			][
				server-send-data remold ['ok module/result]
				if verbose > 0 [log/info ["done in " now/time/precise - t0 " sec"]]
			]
		]
	]
]

if error? set/any 'err try [
	if not encap-fs/cache [ctx-task-class/connect]
][
	change-dir system/options/path
	write/append %crash.log reform [newline now ":" mold disarm :err]
]